/*
 * This is part of XmlEditor, a graphical editor for XML files
 * Copyright (C) 2001 CERN - Sebastien Ponce
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contacts :
 *   sebastien.ponce@polytechnique.org
 *   http://lhcb-comp.web.cern.ch/lhcb-comp/
 *   http://www.cern.ch
 */

package ch.cern.lhcb.xmleditor.transfer;

import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragSourceContext;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceEvent;
import java.io.IOException;
import java.util.Vector;

import ch.cern.lhcb.xmleditor.gui.XmlTree;
import ch.cern.lhcb.xmleditor.specific.IconsManager;
import ch.cern.lhcb.xmleditor.undo.CompoundEdit;
import ch.cern.lhcb.xmleditor.xerces.ElementImpl;
import ch.cern.lhcb.xmleditor.xerces.TreeElement;

/**
 * this is an implementation of the interface XmlDragSourceListener that deals
 * with the drop sources in the XmlEditor. The only source is in fact an XmlTree
 * at this time.
 * 
 * @author Sebastien Ponce
 */
public class DragSourceListener implements java.awt.dnd.DragSourceListener {

    /**
     * the underlying tree
     */
    private XmlTree tree;

    /**
     * the currently used cursor type This is either DnDConstants.ACTION_NONE,
     * DnDConstants.ACTION_MOVE, DnDConstants.ACTION_COPY or
     * DnDConstants.ACTION_LINK
     */
    private int cursorType;

    /**
     * constructor
     * 
     * @param tree the underlying jTree
     */
    public DragSourceListener(XmlTree tree) {
        super();
        this.tree = tree;
        this.cursorType = DnDConstants.ACTION_NONE;
    }

    /**
     * as the hotspot enters a platform dependent drop site
     * 
     * @param dsde The DragSourceDragEvent describing the event that has just
     *        occured
     */
    public void dragEnter(DragSourceDragEvent dsde) {
        this.updateCursor(dsde);
    }

    /**
     * as the hotspot moves over a platform dependent drop site
     * 
     * @param dsde The DragSourceDragEvent describing the event that has just
     *        occured
     */
    public void dragOver(DragSourceDragEvent dsde) {
        this.updateCursor(dsde);
    }

    /**
     * the user has modified the drop gesture
     * 
     * @param dsde The DragSourceDragEvent describing the event that has just
     *        occured
     */
    public void dropActionChanged(DragSourceDragEvent dsde) {
        this.updateCursor(dsde);
    }

    /**
     * as the hotspot exits a platform dependent drop site
     * 
     * @param dse The DragSourceDragEvent describing the event that has just
     *        occured
     */
    public void dragExit(DragSourceEvent dse) {
        setCursor(dse.getDragSourceContext(), cursorType, false);
    }

    /**
     * updates the cursor depending on a DragSourceDragEvent
     * 
     * @param dsde the DragSourceDragEvent
     */
    private void updateCursor(DragSourceDragEvent dsde) {
        // set a cursor, depending on the action and on the location
        // Note the the default action is move, then copy and finally link
        setCursor(dsde.getDragSourceContext(), dsde.getTargetActions(), tree
                .getDraggedRow() != -1);
    }

    /**
     * sets the cursor of a DragSourceContext depending on an action type and a
     * boolean saying if the action is possible or not
     * 
     * @param dsc the DragSourceContext
     * @param action the action type
     * @param ok true if the action is possible
     */
    private void setCursor(DragSourceContext dsc, int action, boolean ok) {
        IconsManager iconsManager = tree.getTreeView().getEditor()
                .getIconsManager();
        if ((action & DnDConstants.ACTION_MOVE) == DnDConstants.ACTION_MOVE) {
            dsc.setCursor(iconsManager.getMoveDropCursor(ok));
            cursorType = DnDConstants.ACTION_MOVE;
        } else if ((action & DnDConstants.ACTION_COPY) == DnDConstants.ACTION_COPY) {
            dsc.setCursor(iconsManager.getCopyDropCursor(ok));
            cursorType = DnDConstants.ACTION_COPY;
        } else if ((action & DnDConstants.ACTION_LINK) == DnDConstants.ACTION_LINK) {
            dsc.setCursor(iconsManager.getLinkDropCursor(ok));
            cursorType = DnDConstants.ACTION_LINK;
        } else {
            // but NoMove by default
            dsc.setCursor(iconsManager.getMoveDropCursor(false));
            cursorType = DnDConstants.ACTION_MOVE;
        }
    }

    /**
     * as the operation completes
     * 
     * @param dsde The DragSourceDragEvent describing the event that has just
     *        occured
     */
    public void dragDropEnd(DragSourceDropEvent dsde) {
        // if the drop operation had no success, does nothing
        if (dsde.getDropSuccess()) {
            // else check at the action type
            int dropAction = dsde.getDropAction();
            // the use of the variable copyOrLinkActionInProgress is only a
            // workaround to a bug
            // found in the sun jvm on windows. This should be removed as soon
            // as this
            // bug get fixed -- FIXME
            if (dropAction == DnDConstants.ACTION_MOVE
                    && !tree.copyOrLinkActionInProgress) {
                // if it was a move action
                // gets the XmlTreeNode transfered and removes the old node
                try {
                    // gets the nodes
                    Vector nodes = (Vector) dsde.getDragSourceContext()
                            .getTransferable().getTransferData(
                                    TransferableNode.LocalFlavor);
                    // begins a compoundedit for undo management since
                    // several nodes are removed at once
                    CompoundEdit innerEdit = new CompoundEdit(tree
                            .getTreeView().getEditor());
                    tree.getTreeView().getEditor().getUndoManager().addEdit(
                            innerEdit);
                    // removes every node
                    for (int i = 0; i < nodes.size(); i++) {
                        TreeElement node = (TreeElement) nodes.elementAt(i);
                        // gets the parent
                        ElementImpl parent = (ElementImpl) node.getParentNode();
                        // removes the node from the document
                        parent.undoableRemoveChild(node);
                    }
                    // closes the compound edit
                    innerEdit.end();
                } catch (IOException io) {
                    // nothing to do since we should never reach this point
                } catch (UnsupportedFlavorException ufe) {
                    // nothing to do since we should never reach this point
                }
            } else {
                // else it was a copy action, does nothing
            }
        }
        // closes the compound edit for the whole dnd action
        tree.getTreeView().getEditor().getUndoManager().endAnyInnerCompound();
    }

}