/*
 * This is part of XmlEditor, a graphical editor for XML files
 * Copyright (C) 2001 CERN - Sebastien Ponce
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contacts :
 *   sebastien.ponce@polytechnique.org
 *   http://lhcb-comp.web.cern.ch/lhcb-comp/
 *   http://www.cern.ch
 */

package ch.cern.lhcb.xmleditor.transfer;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetEvent;
import java.io.IOException;
import java.util.Vector;

import javax.swing.tree.TreePath;

import org.w3c.dom.Node;

import ch.cern.lhcb.xmleditor.gui.MessageDisplayer;
import ch.cern.lhcb.xmleditor.gui.XmlTree;
import ch.cern.lhcb.xmleditor.gui.XmlTreeNode;
import ch.cern.lhcb.xmleditor.undo.CompoundEdit;
import ch.cern.lhcb.xmleditor.xerces.ElementImpl;
import ch.cern.lhcb.xmleditor.xerces.TreeElement;

/**
 * this is an implementation of the interface DropTargetListener that deals with
 * the drop targets in the XmlEditor. The only target is in fact an XmlTree at
 * this time.
 * 
 * @author Sebastien Ponce
 */
public class DropTargetListener implements java.awt.dnd.DropTargetListener {

    /** This is the height of the top border of a cell */
    private final static int CELL_TOP_BORDER_HEIGHT = 4;

    /** This is the height of the bottom border of a cell */
    private final static int CELL_BOTTOM_BORDER_HEIGHT = 4;

    /**
     * the underlying tree
     */
    private XmlTree tree;

    /**
     * the underlying MessageDisplayer
     */
    private MessageDisplayer msgDis;

    /**
     * constructor
     * 
     * @param tree the underlying jTree
     * @param msgDis a messageDisplayer for error display
     */
    public DropTargetListener(XmlTree tree, MessageDisplayer msgDis) {
        super();
        this.tree = tree;
        this.msgDis = msgDis;
    }

    /**
     * accepts or reject the drag on this component depending on the conditions
     * 
     * @param dtde the DropTargetDragEvent concerned
     */
    private void acceptRejectDrag(DropTargetDragEvent dtde) {
        // gets the XmlTreeNode concerned by the drag
        Point p = dtde.getLocation();
        TreePath path = tree.getPathForLocation(p.x, p.y);
        if (path == null) {
            dtde.rejectDrag();
        } else {
            XmlTreeNode node = (XmlTreeNode) path.getLastPathComponent();
            // the only case of rejection is an ACTION_LINK action in case
            // references are not allowed on the target
            if ((dtde.getDropAction() == DnDConstants.ACTION_LINK)
                    && !node.allowsReferences()) {
                dtde.rejectDrag();
            } else {
                dtde.acceptDrag(DnDConstants.ACTION_COPY_OR_MOVE
                        | DnDConstants.ACTION_LINK);
            }
        }
    }

    /**
     * Called when a drag operation has encountered the DropTarget.
     * 
     * @param dtde the DropTargetDragEvent
     */
    public void dragEnter(DropTargetDragEvent dtde) {
        // accepts or reject the drag on this component
        this.acceptRejectDrag(dtde);
    }

    /**
     * Called when a drag operation is ongoing on the DropTarget.
     * 
     * @param dtde the DropTargetDragEvent
     */
    public void dragOver(DropTargetDragEvent dtde) {
        Point p = dtde.getLocation();
        tree.setDraggedRow(tree.getRowForLocation(p.x, p.y));
        // if we are outside a cell
        if (tree.getDraggedRow() == -1) {
            // No drag here
            dtde.rejectDrag();
        } else {
            Rectangle location = tree.getRowBounds(tree.getDraggedRow());
            int fromTop = p.y - location.y;
            int fromBottom = location.height - fromTop;
            if (fromTop < CELL_TOP_BORDER_HEIGHT) {
                tree.setDraggedRowOffset(javax.swing.SwingConstants.TOP);
            } else if (fromBottom < CELL_BOTTOM_BORDER_HEIGHT) {
                tree.setDraggedRowOffset(javax.swing.SwingConstants.BOTTOM);
            } else {
                tree.setDraggedRowOffset(javax.swing.SwingConstants.CENTER);
            }
            // accepts or reject the drag on this component
            this.acceptRejectDrag(dtde);
        }
        // refreshes the display of the tree
        tree.repaint();
    }

    /**
     * Called if the user has modified the current drop gesture.
     * 
     * @param dtde the DropTargetDragEvent
     */
    public void dropActionChanged(DropTargetDragEvent dtde) {
        // accepts or reject the drag on this component
        this.acceptRejectDrag(dtde);
    }

    /**
     * The drag operation has departed the DropTarget without dropping.
     * 
     * @param dte the DropTargetDragEvent
     */
    public void dragExit(DropTargetEvent dte) {
    }

    /**
     * The drag operation has terminated with a drop on this DropTarget.
     * 
     * @param dtde the DropTargetDragEvent
     */
    public void drop(DropTargetDropEvent dtde) {
        try {
            // checks the local flavor is supported
            if (dtde.getTransferable().isDataFlavorSupported(
                    TransferableNode.LocalFlavor)) {
                dropWithLocalFlavor(dtde);
            } else {
                // if no XmlTreeNode, rejects the drop
                dtde.rejectDrop();
            }
        } catch (IOException io) {
            io.printStackTrace();
            dtde.rejectDrop();
        } catch (UnsupportedFlavorException ufe) {
            ufe.printStackTrace();
            dtde.rejectDrop();
        }
    }

    /**
     * The drag operation has terminated with a drop on this DropTarget and the
     * transferable supports the local flavor.
     * 
     * @param dtde the DropTargetDragEvent
     */
    private void dropWithLocalFlavor(DropTargetDropEvent dtde)
            throws IOException, UnsupportedFlavorException {
        // retrieves the row where the drop occured
        int currentRow = tree.getDraggedRow();
        // give up if we are outside any cell
        if (currentRow == -1) {
            dtde.rejectDrop();
            return;
        }
        // retrieves the new parent of the dropped nodes
        // if the drop is done on one of the borders of a node. This means
        // it concerns its parent and the nodes should be placed just before
        // or after this node
        XmlTreeNode parent = (XmlTreeNode) tree.getPathForRow(currentRow)
                .getLastPathComponent();
        if (tree.getDraggedRowOffset() == javax.swing.SwingConstants.TOP
                || tree.getDraggedRowOffset() == javax.swing.SwingConstants.BOTTOM) {
            parent = (XmlTreeNode) parent.getParent();
        }
        // retrieves the nodes to be dropped
        Vector nodes = (Vector) dtde.getTransferable().getTransferData(
                TransferableNode.LocalFlavor);
        // checks that the drop is possible and rejects or accepts it
        if (rejectDropIfNotAllowed(parent, nodes, dtde)) {
            dtde.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE
                    | DnDConstants.ACTION_LINK);
        } else {
            // the drop was rejected, just returns
            return;
        }
        // the use of the variable copyOrLinkActionInProgress is only a
        // workaround to a bug
        // found in the sun jvm on windows. This should be removed as soon as
        // this
        // bug get fixed -- FIXME
        if (dtde.getDropAction() == DnDConstants.ACTION_COPY
                || dtde.getDropAction() == DnDConstants.ACTION_LINK) {
            tree.copyOrLinkActionInProgress = true;
        } else {
            tree.copyOrLinkActionInProgress = false;
        }
        // clears the selection in the tree
        tree.clearSelection();
        // calls the actual drop method
        boolean success = internalDropWithLocalFlavor((ElementImpl) parent
                .getNode(), nodes, dtde.getDropAction());
        // validates the drop
        dtde.getDropTargetContext().dropComplete(success);
    }

    /**
     * Drops nodes into parent with action action. Does no check at all.
     * 
     * @param parent the node where to drop
     * @param nodes a vector of TreeElements to be dropped
     * @param action the action for this drop
     * @return true is drop was successfull
     */
    private boolean internalDropWithLocalFlavor(ElementImpl parent,
            Vector nodes, int action) {
        // gets the neighbour for this drop
        Node neighbour = this.getNeighbour();
        // begins a compoundedit for undo management since several nodes
        // may be pasted at once
        CompoundEdit edit = new CompoundEdit(tree.getTreeView().getEditor());
        tree.getTreeView().getEditor().getUndoManager().addEdit(edit);
        // clones the nodes and pastes them one by one
        boolean success = true;
        for (int i = 0; i < nodes.size(); i++) {
            TreeElement node = (TreeElement) nodes.elementAt(i);
            TreeElement newNode = (TreeElement) node.cloneNode(true);
            // this boolean tells whether we want the tree to be redisplayed
            // after the paste. This is only the case for the last node and if
            // the action is not MOVE. Otherwise, it will be redisplayed later
            // (when the deletion will occur in case of move action)
            boolean redisplay = (action != DnDConstants.ACTION_MOVE)
                    && (i == nodes.size() - 1);
            // pastes the node at the right position
            if (action == DnDConstants.ACTION_COPY
                    || action == DnDConstants.ACTION_MOVE) {
                // MOVE and COPY actions
                if (tree.getDraggedRowOffset() == javax.swing.SwingConstants.TOP) {
                    // if before neighbour
                    success = success
                            && parent.insertNodeBefore(newNode, neighbour,
                                    redisplay);
                } else if (tree.getDraggedRowOffset() == javax.swing.SwingConstants.BOTTOM) {
                    // if after neighbour
                    success = success
                            && parent.insertNodeAfter(newNode, neighbour,
                                    redisplay);
                } else {
                    // if inside parent
                    success = success && parent.pasteNode(newNode, redisplay);
                }
            } else if (action == DnDConstants.ACTION_LINK) {
                // LINK action : pastes a reference on the node
                if (tree.getDraggedRowOffset() == javax.swing.SwingConstants.TOP) {
                    // if before neighbour
                    success = success
                            && parent.insertRefNodeBefore(node, neighbour, tree
                                    .getTreeView(), redisplay);
                } else if (tree.getDraggedRowOffset() == javax.swing.SwingConstants.BOTTOM) {
                    // if after neighbour
                    success = success
                            && parent.insertRefNodeAfter(node, neighbour, tree
                                    .getTreeView(), redisplay);
                } else {
                    // if inside parent
                    success = success
                            && parent.pasteRefNode(node, tree.getTreeView(),
                                    redisplay);
                }
            }
        }
        // closes the compound edit
        edit.end();
        // returns
        return success;

    }

    /**
     * gets the neighbour of nodes being currently dropped or returns null if
     * there is none. The definition of the neighbour is the following : when
     * dropping a node, the drop can be done on another node or on one of the
     * borders of another node. In the first case, there is no neighbour for the
     * drop. In the second case, the drop should be done in the node's parent,
     * just before or after the node. This node is called the neighbour
     * 
     * @return the neighbour of nodes being currently dropped
     */
    private Node getNeighbour() {
        // retrieves the neighbour of the dropped nodes
        XmlTreeNode neighbour = null;
        if (tree.getDraggedRowOffset() == javax.swing.SwingConstants.TOP
                || tree.getDraggedRowOffset() == javax.swing.SwingConstants.BOTTOM) {
            neighbour = (XmlTreeNode) tree.getPathForRow(tree.getDraggedRow())
                    .getLastPathComponent();
        }
        if (neighbour != null) {
            // take care if the neighbour is a opened reference
            // it should then be replaced by the reference itself since the
            // node will be pasted into the file where the reference is
            if (neighbour.getOldNode() != null) {
                neighbour = neighbour.getOldNode();
            }
            // returns the neighbour as a node
            if (neighbour != null) { return neighbour.getNode(); }
        }
        return null;
    }

    /**
     * Checks whether a drop of nodes into parent is possible or not and rejects
     * the drop if not. This check includes that parent is a container for nodes
     * and that no recursive tree is being built
     * 
     * @param parent the node where to drop
     * @param nodes a vector of TreeElements to be dropped
     * @param dtde the DropTargetDropEvent corresponding to the drop
     * @return true is drop is possible and was not rejected
     */
    private boolean rejectDropIfNotAllowed(XmlTreeNode parent, Vector nodes,
            DropTargetDropEvent dtde) {
        // checks we're pasting into an element
        if (!(parent.getNode() instanceof ElementImpl)) {
            dtde.rejectDrop();
            msgDis.popupMessage("Drop canceled");
            msgDis.displayMessage("You can only drop inside an element.\n");
            return false;
        }
        // if references are not allowed, reject any link drop
        if (!parent.allowsReferences()
                && ((dtde.getDropAction() & DnDConstants.ACTION_LINK) == DnDConstants.ACTION_LINK)) {
            msgDis
                    .displayMessage("There is no sense in creating a link to a node !\n");
            return false;
        }
        // checks we are not pasting into a reference
        if (parent.isReference()) {
            dtde.rejectDrop();
            msgDis.popupMessage("Drop canceled");
            msgDis
                    .displayMessage("You cannot drop a node inside a reference.\n");
            return false;
        }
        // check we are not moving a node inside itself.
        // Copying or referencing a node inside itself is however
        // allowed
        if ((dtde.getDropAction() & DnDConstants.ACTION_MOVE) == DnDConstants.ACTION_MOVE) {
            boolean noRecursivity = true;
            for (int i = 0; (i < nodes.size()) && noRecursivity; i++) {
                TreeElement node = (TreeElement) nodes.elementAt(i);
                if (parent.isNodeAncestor(node)) {
                    noRecursivity = false;
                }
            }
            if (!noRecursivity) {
                dtde.rejectDrop();
                msgDis.popupMessage("Drop canceled");
                msgDis
                        .displayMessage("One of the nodes you tried to drop is an "
                                + "ancestor of the node into which you are "
                                + "dropping.\nSince you requested a move, this "
                                + "drop has no sense. I cannot move a node "
                                + "inside itself !\nMaybe you meant to copy "
                                + "or reference it. Try to hit Ctrl or Ctrl + "
                                + "Shift while dropping in this case.\n");
                return false;
            }
        }
        return true;
    }

}
