/*
 * This is part of XmlEditor, a graphical editor for XML files
 * Copyright (C) 2001 CERN - Sebastien Ponce
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contacts :
 *   sebastien.ponce@polytechnique.org
 *   http://lhcb-comp.web.cern.ch/lhcb-comp/
 *   http://www.cern.ch
 */

package ch.cern.lhcb.xmleditor.xerces;

import java.util.HashSet;
import java.util.Iterator;

import javax.swing.tree.TreePath;
import javax.swing.undo.UndoableEdit;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import ch.cern.lhcb.xmleditor.gui.MessageDisplayer;
import ch.cern.lhcb.xmleditor.gui.XmlMouseListener;
import ch.cern.lhcb.xmleditor.gui.XmlTree;
import ch.cern.lhcb.xmleditor.gui.XmlTreeNode;
import ch.cern.lhcb.xmleditor.specific.Model;
import ch.cern.lhcb.xmleditor.specific.XmlDisplay;
import ch.cern.lhcb.xmleditor.specific.XmlFile;
import ch.cern.lhcb.xmleditor.undo.CompoundEdit;
import ch.cern.lhcb.xmleditor.undo.UndoManager;
import ch.cern.lhcb.xmleditor.undo.UndoableAttributeEdition;
import ch.cern.lhcb.xmleditor.undo.UndoableElementAddition;
import ch.cern.lhcb.xmleditor.undo.UndoableElementRemoval;

/**
 * This class subclasses the default ElementImpl implementation to take some
 * displaying facts into account
 * 
 * @author Sebastien Ponce
 */
public class ElementImpl extends org.apache.xerces.dom.ElementImpl implements
        TreeElement {

    /**
     * this is the set of TreeNodes associated with this element
     */
    private HashSet treeNodes;

    /**
     * constructor
     * 
     * @param ownerDoc the document that created it
     * @param name the name of the new Element
     */
    public ElementImpl(DocumentImpl ownerDoc, String name) {
        super(ownerDoc, name);
        treeNodes = new HashSet();
    }

    /**
     * Returns a duplicate of this node.
     * 
     * @param deep if true, performs a deep copy
     */
    public Node cloneNode(boolean deep) {
        // this trick is here to avoid the new node to use the TreeNodes of
        // the old one. The problem is that children creation in the new node
        // would in such a case end up with creation of new TreeNodes
        HashSet temp = treeNodes;
        treeNodes = new HashSet();
        ElementImpl result = (ElementImpl) super.cloneNode(deep);
        treeNodes = temp;
        return result;
    }

    /**
     * adds treeNode to the list of treeNode associated with this node
     * 
     * @param treeNode the TreeNode to add to the list
     */
    public void addTreeNode(XmlTreeNode treeNode) {
        treeNodes.add(treeNode);
        // also adds the new treeNode to all attributes of this node
        NamedNodeMap list = this.getAttributes();
        for (int i = 0; i < list.getLength(); i++) {
            ((AttrImpl) list.item(i)).addTreeNode(treeNode);
        }
    }

    /**
     * removes treeNode from the list of treeNode associated with this node
     * 
     * @param treeNode the TreeNode to remove from the list
     */
    public void removeTreeNode(XmlTreeNode treeNode) {
        treeNodes.remove(treeNode);
        // also removes the new treeNode from all attributes of this node
        NamedNodeMap list = this.getAttributes();
        for (int i = 0; i < list.getLength(); i++) {
            ((AttrImpl) list.item(i)).removeTreeNode(treeNode);
        }
    }

    /**
     * updates the upToDate flag of the file corresponding to this node
     */
    private void updateFile() {
        XmlFile file = ((DocumentImpl) this.getOwnerDocument()).getXmlFile();
        if (file != null) {
            file.upToDate = false;
        }
    }

    /**
     * this method should be called whenever an attribute is added to this
     * element
     * 
     * @param name the name of the new attribute
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     */
    public void attrAdded(String name, boolean redisplay) {
        // updates the upToDate field of the corresponding file
        this.updateFile();
        // gets the new attribute's node
        AttrImpl attr = (AttrImpl) this.getAttributeNode(name);
        // for every representation of this element
        Iterator iterator = treeNodes.iterator();
        while (iterator.hasNext()) {
            XmlTreeNode treeNode = (XmlTreeNode) iterator.next();
            // the representation of the parent must be copied into the child
            attr.addTreeNode(treeNode);
            // redisplays the whole tree to validate the changes if needed
            if (redisplay) {
                treeNode.getXmlTree().getTreeView().redisplayAll();
            }
        }
    }

    /**
     * this method should be called whenever a child is added to this element
     * 
     * @param node the added child
     * @param refNode the child before which the new child was inserted
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     */
    public void childAdded(TreeElement node, Node refNode, boolean redisplay) {
        // updates the upToDate field of the corresponding file
        this.updateFile();
        // for each representation of the parent,
        // adds a representation of the child
        Iterator iterator = treeNodes.iterator();
        while (iterator.hasNext()) {
            // gets the treeNode
            XmlTreeNode treeNode = (XmlTreeNode) iterator.next();
            // gets the corresponding tree
            XmlTree tree = treeNode.getXmlTree();
            // gets the corresponding model
            Model model = tree.getXmlModel();
            // the index of the new node among the children
            int index = treeNode.getChildCount();
            if (refNode != null) {
                // else, put it just before the reference node
                // or just before the first node after this one which
                // is displayed or at the end of the children
                Node sibling = refNode;
                int maybeIndex = -1;
                while (maybeIndex == -1 && sibling != null) {
                    maybeIndex = treeNode.getChildIndex(sibling);
                    sibling = sibling.getNextSibling();
                }
                if (maybeIndex > -1) {
                    index = maybeIndex;
                }
            }
            // adds the node to the tree of displayed ones
            XmlTreeNode childTreeNode = model.insertTree(node, treeNode, index);
            // adds treeNode to the list of representations for this node
            node.addTreeNode(childTreeNode);
            // if node is displayed but not visible (ie its parent is not
            // expanded), makes it visible
            XmlDisplay xmlDisplay = getXmlDisplay();
            boolean displayed;
            if (node.getNodeType() == Node.TEXT_NODE) {
                displayed = !tree.getTreeView().isValueDisplayedAsAttribute();
            } else if (node.getNodeName() == XmlMouseListener.COMMENT) {
                displayed = tree.getTreeView().isCommentDisplayed();
            } else {
                if (xmlDisplay != null) {
                    displayed = xmlDisplay.isSubElementDisplayed(this
                            .getNodeName(), node.getNodeName());
                } else {
                    displayed = true;
                }
            }
            if (displayed) {
                TreePath path = (new TreePath(treeNode.getPath()))
                        .pathByAddingChild(treeNode.getChildAt(index));
                tree.makeVisible(path);
                tree.scrollPathToVisible(path);
                tree.setSelectionPath(path);
            }
            // redisplays the whole tree to validate the changes
            if (redisplay) {
                tree.getTreeView().redisplayAll();
            }
        }
    }

    /**
     * this method should be called when one representation of this node is
     * removed
     * 
     * @param treeNode the representation to be removed. No check done on this
     *        parameter. It MUST be non-null and representing this node
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     */
    private void internalRepresentationRemoved(XmlTreeNode treeNode,
            boolean redisplay) {
        // gets the corresponding tree
        XmlTree tree = treeNode.getXmlTree();
        // gets the corresponding model
        Model model = tree.getXmlModel();
        // notice if the node was selected
        boolean selected = tree
                .isPathSelected(new TreePath(treeNode.getPath()));
        // removes the TreeNode from the tree
        model.removeNodeFromParent(treeNode);
        // removes treeNode from the list of representations for this node
        this.removeTreeNode(treeNode);
        // if the node was selected, re-initializes the editor Pane
        if (selected) {
            tree.setSelectionPath(null);
            tree.getTreeView().eraseEditorPane();
        }
        // redisplays the whole tree to validate the changes if needed
        if (redisplay) {
            tree.getTreeView().redisplayAll();
        }
    }

    /**
     * this method should be called when this node is removed
     */
    public void removed() {
        // updates the upToDate field of the corresponding file
        this.updateFile();
        // for each representation of the parent
        while (!treeNodes.isEmpty()) {
            // gets the treeNode
            XmlTreeNode treeNode = (XmlTreeNode) treeNodes.iterator().next();
            // removes it
            internalRepresentationRemoved(treeNode, true);
        }
    }

    /**
     * this method should be called whenever the name of this node is edited
     */
    public void nodeNameChanged() {
        // updates the upToDate field of the corresponding file
        this.updateFile();
        // updates the different displays of this node
        Iterator iterator = treeNodes.iterator();
        while (iterator.hasNext()) {
            XmlTreeNode treeNode = (XmlTreeNode) iterator.next();
            // redisplays the whole tree to validate the changes
            treeNode.getXmlTree().getTreeView().redisplayAll();
        }
    }

    /**
     * this method should be called whenever the value of this node is edited
     */
    public void nodeValueChanged() {
        // updates the upToDate field of the corresponding file
        this.updateFile();
        // updates the different displays of this node
        Iterator iterator = treeNodes.iterator();
        while (iterator.hasNext()) {
            XmlTreeNode treeNode = (XmlTreeNode) iterator.next();
            // redisplays the whole tree to validate the changes
            treeNode.getXmlTree().getTreeView().redisplayAll();
        }
    }

    /**
     * get the first attribute of this node with the given name
     * 
     * @return the first attribute whose name is name or null if there are none
     */
    public Node getAttributeByName(String name) {
        // looks at every attribute and checks the names
        NamedNodeMap nnm = this.getAttributes();
        if (nnm == null) { return null; }
        for (int i = 0; i < nnm.getLength(); i++) {
            Node node = nnm.item(i);
            if (node.getNodeName().equals(name)) { return node; }
        }
        // if nothing was found, return null
        return null;
    }

    /**
     * returns the first text subelement of this node
     * 
     * @return its first text subelement or null if there is none
     */
    public Node getPCDATA() {
        // retrieves the children of the node
        NodeList nodeList = this.getChildNodes();
        // for each child
        for (int i = 0; i < nodeList.getLength(); i++) {
            // retrieves the child of the node
            Node child = nodeList.item(i);
            // checks its type
            if (child.getNodeType() == Node.TEXT_NODE) { return child; }
        }
        // nothing found
        return null;
    }

    /**
     * gets the value of this node, if it has one. The value of such a node is
     * in fact the value of its text child if it exists
     * 
     * @return the value of this node, null it has none
     */
    public String getValue() {
        Node valueNode = this.getPCDATA();
        if (valueNode != null) {
            return valueNode.getNodeValue();
        } else {
            return null;
        }
    }

    /**
     * gets the value of the first attribute of this node with name name, if it
     * exists.
     * 
     * @param name the name of the attribute whose value we are looking for
     * @return the value of the first attribute with name name or, null there is
     *         none
     */
    public String getAttributeValue(String name) {
        Node valueNode = this.getAttributeByName(name);
        if (valueNode != null) {
            return valueNode.getNodeValue();
        } else {
            return null;
        }
    }

    /**
     * sets the value of this node, if it has one. The value of such a node is
     * in fact the value of its text child if it exists
     * 
     * @param value the new value of this node
     */
    public void setValue(String value) {
        this.getPCDATA().setNodeValue(value);
    }

    /**
     * says whether node is an ancestor of this node or not Take care that the
     * notion of ancestor does include the parent-child relation as well as the
     * element-attribute one. This means that an element is ancestor of its
     * attributes.
     * 
     * @param node the node to be tested
     * @return true if the current node is a descendant of node
     */
    public boolean isNodeAncestor(TreeElement node) {
        if (node == this) { return true; }
        TreeElement parent = (TreeElement) this.getParentNode();
        if (parent == null) {
            return false;
        } else {
            return parent.isNodeAncestor(node);
        }
    }

    /**
     * gets the XmlDisplay for this node
     * 
     * @return the XmlDisplay for this node
     */
    private XmlDisplay getXmlDisplay() {
        XmlFile file = ((DocumentImpl) this.getOwnerDocument()).getXmlFile();
        XmlDisplay xmlDisplay = null;
        if (file != null) {
            xmlDisplay = file.getXmlDisplay();
        }
        return xmlDisplay;
    }

    /**
     * creates a new attribute of type type as child of this element
     * 
     * @param type the type of the node to be created
     * @param msgDis a MessageDisplayer for error display
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     */
    public void createAttribute(String type, MessageDisplayer msgDis,
            boolean redisplay) {
        // checks if such an attribute already exists.
        // If yes, don't create one more
        if (this.getAttributes().getNamedItem(type) != null) {
            msgDis.popupMessage("Attribute creation canceled");
            msgDis
                    .displayMessage("The current node already contains an attribute "
                            + "of this type.\nIf you really want to replace it "
                            + "by a new one instead of just changing the value, "
                            + "please remove the old one before creating a new one.\n");
            return;
        }
        // gets the XmlDisplay if it exists
        XmlDisplay xmlDisplay = getXmlDisplay();
        // tries to get a default value for this attribute
        String value;
        if (xmlDisplay != null) {
            value = xmlDisplay.getAttributeDefaultValue(this.getNodeName(),
                    type);
        } else {
            value = "aValue";
        }
        // creates the new attribute
        this.undoableSetAttribute(type, value, redisplay);
    }

    /**
     * removes an attribute (or an element displayed as attribute) from this
     * element
     * 
     * @param name the name of the attribute to be removed
     * @param msgDis a MessageDisplayer for error display
     */
    public void removeAttribute(String name, MessageDisplayer msgDis) {
        // check whether it is an attribute
        if (this.getAttributes().getNamedItem(name) != null) {
            // removes the attribute from the document
            this.undoableRemoveAttribute(name);
            // if the attribute had a default value, it has not really been
            // removed but changed. Lets the user know it
            if (this.getAttributes().getNamedItem(name) != null) {
                msgDis
                        .popupMessage("Attribute was recreated with default value");
                msgDis
                        .displayMessage("The attribute you removed is known to have a default "
                                + "value.\n"
                                + "As a consequence, it was recreated with this one.\n");
            }
        } else {
            // else it should be an element displayed as attribute
            // finds it
            NodeList list = this.getChildNodes();
            for (int i = 0; i < list.getLength(); i++) {
                Node child = list.item(i);
                // if it's the right one, removes it
                if (child.getNodeName().equals(name)
                        || (child.getNodeName().equals("#text") && name
                                .equals("Value"))) {
                    // for some tricky reason, the subelement has to be
                    // displayed to be removed so we try to know if it was and
                    // do what is needed. This is actually due to the
                    // implementation of XmlTreeNode that hides the non
                    // displayed nodes. See this class for more details
                    XmlDisplay xmlDisplay = getXmlDisplay();
                    boolean displayed = true;
                    if (xmlDisplay != null) {
                        displayed = xmlDisplay.isSubElementDisplayed(this
                                .getNodeName(), name);
                    }
                    if (!displayed) {
                        // display the subelement
                        xmlDisplay.setSubElementDisplayed(this.getNodeName(),
                                name, true);
                    }
                    // removes the node from the document
                    this.undoableRemoveChild(child);
                    if (!displayed) {
                        // hide the subelement that was displayed for its
                        // removing
                        xmlDisplay.setSubElementDisplayed(this.getNodeName(),
                                name, false);
                    }
                    return;
                }
            }
            // else let the user know that no attribute was deleted
            msgDis.popupMessage("Attribute did not exit");
            msgDis
                    .displayMessage("This attribute you removed does actually not exist on "
                            + "this node.\nAs a consequence, it was not deleted :-).\n");
        }
    }

    /**
     * creates a new node of type type and sets default values or attributes
     * 
     * @param type the type of the node to be created
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     * @return the node created
     */
    private ElementImpl internalCreateNode(String type, boolean redisplay) {
        // creates the new node
        ElementImpl newNode = (ElementImpl) this.getOwnerDocument()
                .createElement(type);
        // gets the XmlDisplay if it exists
        XmlDisplay xmlDisplay = getXmlDisplay();
        if (xmlDisplay != null) {
            // if its value is PCDATA, creates an empty text subnode
            if (xmlDisplay.isElementAttributeLike(type)) {
                Text text = this.getOwnerDocument().createTextNode("aText");
                newNode.undoableInsertBeforeChild(text, null, redisplay);
            }
            // tries to add the required attributes
            String[] attributeNames = xmlDisplay
                    .getRequiredAttributeNames(type);
            for (int i = 0; i < attributeNames.length; i++) {
                // tries to get a default value for this attribute
                String value = xmlDisplay.getAttributeDefaultValue(type,
                        attributeNames[i]);
                // creates the new attribute
                newNode.undoableSetAttribute(attributeNames[i], value,
                        redisplay);
            }
        }
        // returns
        return newNode;
    }

    /**
     * creates a new node of type type as child of this element
     * 
     * @param type the type of the node to be created
     * @param treeNode the XmlTreeNode on which the user worked for creating a
     *        node
     * @param msgDis a MessageDisplayer for error display
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     */
    public void createNode(String type, XmlTreeNode treeNode,
            MessageDisplayer msgDis, boolean redisplay) {
        // begins a compoundedit for undo management since the creation of a
        // node can have several steps (in case of default attributes for
        // example)
        UndoManager undoManager = ((DocumentImpl) this.getOwnerDocument())
                .getUndoManager();
        CompoundEdit edit = new CompoundEdit(undoManager.getEditor());
        this.addEdit(edit);
        // creates the node
        ElementImpl newNode = this.internalCreateNode(type, redisplay);
        // adds the new node to its parent
        this.undoableInsertBeforeChild(newNode, null, true);
        // closes the compound edit
        edit.end();
        // if the new node is not displayed, issues a warning
        XmlDisplay xmlDisplay = getXmlDisplay();
        if (xmlDisplay != null) {
            if (!xmlDisplay.isSubElementDisplayed(this.getNodeName(), type)
                    && !xmlDisplay.isSubElementDisplayedAsAttr(this
                            .getNodeName(), type)) {
                msgDis.popupMessage("Node was created but not displayed");
                msgDis
                        .displayMessage("The new node is not currently displayed due to the "
                                + "display configuration but it was successfully "
                                + "created.\n");
            }
        }
    }

    /**
     * creates a new comment node as child of this element
     * 
     * @param tree the XmlTree where the node was created
     * @param treeNode the XmlTreeNode on which the user worked for creating a
     *        node
     * @param msgDis a MessageDisplayer for error display
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     */
    public void createComment(XmlTree tree, XmlTreeNode treeNode,
            MessageDisplayer msgDis, boolean redisplay) {
        // creates the new node
        TreeElement newNode = (TreeElement) this.getOwnerDocument()
                .createComment("aComment");
        // adds the new node to its parent
        this.undoableInsertBeforeChild(newNode, null, redisplay);
        // if the new node is not displayed, issues a warning
        if (!tree.getTreeView().isCommentDisplayed()) {
            msgDis.popupMessage("Comment was created but not displayed");
            msgDis
                    .displayMessage("The new comment is not currently displayed due to the "
                            + "display configuration but it was successfully created.\n");
        }
    }

    /**
     * creates a new text node as child of this element
     * 
     * @param tree the XmlTree where the node was created
     * @param treeNode the XmlTreeNode on which the user worked for creating a
     *        node
     * @param msgDis a MessageDisplayer for error display
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     */
    public void createValue(XmlTree tree, XmlTreeNode treeNode,
            MessageDisplayer msgDis, boolean redisplay) {
        // creates the new node
        TreeElement newNode = (TreeElement) this.getOwnerDocument()
                .createTextNode("aTextNode");
        // adds the new node to its parent
        this.undoableInsertBeforeChild(newNode, null, redisplay);
        // if the new node is not displayed, issues a warning
        if (tree.getTreeView().isValueDisplayedAsAttribute()) {
            msgDis
                    .displayMessage("The new text node is not currently displayed in the tree "
                            + "due to the display configuration but it was successfully "
                            + "created and is displayed as an attribute.\n");
        }
    }

    /**
     * says whether this node is a reference or not. Being a reference only
     * means here having a href attribute.
     * 
     * @return true if this node is a reference
     */
    public boolean isReference() {
        return this.getAttributeByName("href") != null;
    }

    /**
     * Removes the child node indicated by oldChild from the list of children,
     * and returns it. This overrides the default implementation in order to
     * take display into account.
     * 
     * @param oldChild the child to be removed
     * @return the removed child
     */
    public Node removeChild(Node oldChild) throws DOMException {
        // removes the node
        Node result = super.removeChild(oldChild);
        // updates display
        if (result instanceof TreeElement) {
            ((TreeElement) result).removed();
        }
        return result;
    }

    /**
     * Adds a child node to the end of the list of children for this node.
     * Convenience shorthand for insertBefore(newChild,null). This overrides the
     * default implementation in order to take display into account.
     * 
     * @param newChild The node to add
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     * @return the node added
     */
    public Node appendChild(Node newChild, boolean redisplay)
            throws DOMException {
        return this.insertBefore(newChild, null, redisplay);
    }

    /**
     * Adds the node newChild to the end of the list of children of this node.
     * This overrides the default implementation in order to take display into
     * account.
     * 
     * @param newChild the node to add.
     * @param refChild the child before which the new child will be inserted
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     * @return the node added
     */
    public Node insertBefore(Node newChild, Node refChild, boolean redisplay)
            throws DOMException {
        // appends the child
        Node result = super.insertBefore(newChild, refChild);
        // updates display
        // here, we use refChild, since it is the displayed node
        if (result instanceof TreeElement) {
            this.childAdded((TreeElement) result, refChild, redisplay);
        }
        // returns
        return result;
    }

    /**
     * Removes an attribute by name. This overrides the default implementation
     * in order to take display into account.
     * 
     * @param name the name of the attribute to remove
     */
    public void removeAttribute(java.lang.String name) throws DOMException {
        Node attr = this.getAttributeNode(name);
        // removes the attribute
        super.removeAttribute(name);
        // updates display
        if (attr instanceof TreeElement) {
            ((TreeElement) attr).removed();
        }
    }

    /**
     * Adds a new attribute. This overrides the default implementation in order
     * to take display into account.
     * 
     * @param name The name of the attribute to create or alter.
     * @param value Value to set in string form.
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     */
    public void setAttribute(java.lang.String name, java.lang.String value,
            boolean redisplay) throws DOMException {
        // sets the attribute
        super.setAttribute(name, value);
        // updates display
        this.attrAdded(name, redisplay);
    }

    /**
     * Removes the child node indicated by oldChild from the list of children,
     * and returns it. Makes the undoManager aware of this action
     * 
     * @param oldChild the child to be removed
     * @return the removed child
     */
    public Node undoableRemoveChild(Node oldChild) throws DOMException {
        // keep the node following oldChild
        Node refChild = oldChild.getNextSibling();
        // removes the node
        Node result = this.removeChild(oldChild);
        // stores the new action in the UndoManager
        this.addEdit(new UndoableElementRemoval(this.getOwnerDocument(),
                (TreeElement) oldChild, this, (TreeElement) refChild));
        return result;
    }

    /**
     * Adds the node newChild to the list of children of this node, before node
     * refNode. Makes the undoManager aware of this action
     * 
     * @param newChild - The node to add.
     * @param refChild - The node before which the new child should be added.
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     * @return the node added
     */
    private Node undoableInsertBeforeChild(Node newChild, Node refChild,
            boolean redisplay) throws DOMException {
        // appends the child
        Node result = this.insertBefore(newChild, refChild, redisplay);
        // stores the new action in the UndoManager
        this.addEdit(new UndoableElementAddition(this.getOwnerDocument(),
                (TreeElement) newChild, this, (TreeElement) refChild));
        // returns
        return result;
    }

    /**
     * Removes an attribute by name. Makes the undoManager aware of this action
     * 
     * @param name the name of the attribute to remove
     */
    public void undoableRemoveAttribute(java.lang.String name)
            throws DOMException {
        Attr attr = this.getAttributeNode(name);
        if (attr != null) {
            // stores the new action in the UndoManager
            this.addEdit(new UndoableAttributeEdition(this.getOwnerDocument(),
                    name, attr.getValue(), null, this, "remove attribute "
                            + name));
            // removes the attribute
            this.removeAttribute(name);
        }
    }

    /**
     * Adds a new attribute. Makes the undoManager aware of this action
     * 
     * @param name The name of the attribute to create or alter.
     * @param value Value to set in string form.
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     */
    public void undoableSetAttribute(java.lang.String name,
            java.lang.String value, boolean redisplay) throws DOMException {
        // gets the ols value of the attribute
        // there is a little processing since we don't have the same value
        // for a non defined attribute than xerces
        String oldValue = this.getAttribute(name);
        if (oldValue.equals("")) {
            oldValue = null;
        }
        
        // sets the attribute
        this.setAttribute(name, value, redisplay);
        // stores the new action in the UndoManager
        this.addEdit(new UndoableAttributeEdition(this.getOwnerDocument(),
                name, oldValue, value, this, "set attribute " + name));
    }

    /**
     * this adds an edition object to the UndoManager. The UndoManager is
     * retrieved from the ownerDocument. If no UndoManager is available, nothing
     * is done but no exception is raised.
     * 
     * @param edit the edition to add
     * @return true if successfull
     */
    private boolean addEdit(UndoableEdit edit) {
        UndoManager undoManager = ((DocumentImpl) this.getOwnerDocument())
                .getUndoManager();
        if (undoManager != null) { return undoManager.addEdit(edit); }
        return false;
    }

    /**
     * tries to paste the XmlTreeNode node as child of this node. Takes care
     * that this paste does not create a cyle in the node tree
     * 
     * @param node the node to be pasted
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     * @return true if the paste was successfull
     */
    public boolean pasteNode(TreeElement node, boolean redisplay) {
        return this.insertNodeBefore(node, null, redisplay);
    }

    /**
     * tries to insert the XmlTreeNode node as child of this node, after node
     * refNode.. Takes care that this paste does not create a cyle in the node
     * tree
     * 
     * @param node the node to be pasted
     * @param refNode the node after which the new node will be inserted
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     * @return true if the paste was successfull
     */
    public boolean insertNodeAfter(TreeElement node, Node refNode,
            boolean redisplay) {
        return this.insertNodeBefore(node, refNode.getNextSibling(), redisplay);
    }

    /**
     * tries to insert the XmlTreeNode node as child of this node, before node
     * refNode.. Takes care that this paste does not create a cyle in the node
     * tree
     * 
     * @param node the node to be pasted
     * @param refNode the node before which the new node will be inserted
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     * @return true if the paste was successfull
     */
    public boolean insertNodeBefore(TreeElement node, Node refNode,
            boolean redisplay) {
        // the node to actually put into the tree
        TreeElement newNode = buildNodeForInsertion(node);
        // adds the new node to its this
        this.undoableInsertBeforeChild(newNode, refNode, redisplay);
        // all was fine, so return true
        return true;
    }

    /**
     * builds a node for insertion in this document. This includes importing the
     * node if necessary.
     * 
     * @param node the node to be inserted
     * @return the built node
     */
    private TreeElement buildNodeForInsertion(TreeElement node) {
        if (this.getOwnerDocument() == node.getOwnerDocument()) {
            // if the new node is dropped in the old node document
            // just duplicates the node here
            return node;
        } else {
            // else the new node does not belong to the same document
            // as the old one
            // imports the node into the new document
            return (TreeElement) this.getOwnerDocument().importNode(node, true);
        }
    }

    /**
     * creates a new reference node pointing to node and inserts it into this
     * document at the end of the ndoe list
     * 
     * @param node the node that will be referenced
     * @param msgDis a MessageDisplayer for error display
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     * @return true if creation was successfull
     */
    public boolean pasteRefNode(TreeElement node, MessageDisplayer msgDis,
            boolean redisplay) {
        return this.insertRefNodeBefore(node, null, msgDis, redisplay);
    }

    /**
     * creates a new reference node pointing to node and inserts it into this
     * document before node refNode
     * 
     * @param node the node that will be referenced
     * @param refNode the node before which the new node will be inserted
     * @param msgDis a MessageDisplayer for error display
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     * @return true if creation was successfull
     */
    public boolean insertRefNodeAfter(TreeElement node, Node refNode,
            MessageDisplayer msgDis, boolean redisplay) {
        return this.insertRefNodeBefore(node, refNode.getNextSibling(), msgDis,
                redisplay);
    }

    /**
     * creates a new reference node pointing to node and inserts it into this
     * document before node refNode
     * 
     * @param node the node that will be referenced
     * @param refNode the node before which the new node will be inserted
     * @param msgDis a MessageDisplayer for error display
     * @param redisplay this tells whether to redisplay everything after the
     *        change is done
     * @return true if creation was successfull
     */
    public boolean insertRefNodeBefore(TreeElement node, Node refNode,
            MessageDisplayer msgDis, boolean redisplay) {
        // tries to build the reference
        String ref = buildNodeReference(node, msgDis);
        if (ref == null) { return false; }
        // begins a compoundedit for undo management since the creation of a
        // reference has several steps
        UndoManager undoManager = ((DocumentImpl) this.getOwnerDocument())
                .getUndoManager();
        CompoundEdit edit = new CompoundEdit(undoManager.getEditor());
        this.addEdit(edit);
        // creates the node
        ElementImpl newNode = this.internalCreateNode(node.getNodeName()
                + "ref", redisplay);
        // adds the new node to its parent
        this.undoableInsertBeforeChild(newNode, refNode, redisplay);
        // set the reference
        newNode.undoableSetAttribute("href", ref, redisplay);
        // closes the compound edit
        edit.end();
        return true;
    }

    /**
     * builds the reference on Node node. Returns null if it is not possible to
     * reference node in this element
     * 
     * @param node the node to reference
     * @param msgDis a MessageDisplayer for error display
     * @return the reference
     */
    private String buildNodeReference(TreeElement node, MessageDisplayer msgDis) {
        // avoid references on references
        if (node.getNodeName().endsWith("ref")) {
            msgDis.popupMessage("No reference created");
            msgDis
                    .displayMessage("The node you tried to create a reference on was a "
                            + "reference itself.\n"
                            + "Since references on references are not allowed, you "
                            + "cannot create it.\n");
            return null;
        }
        // avoid references on node that are no element
        if (!(node instanceof ElementImpl)) {
            msgDis.popupMessage("No reference created");
            msgDis
                    .displayMessage("You can only create a reference on elements. "
                            + "This was not an element.\n");
            return null;
        }
        // gets the file where the node will be put, ie the file of this element
        XmlFile parentFile = ((DocumentImpl) this.getOwnerDocument())
                .getXmlFile();
        // gets the file from which the node comes
        XmlFile nodeFile = ((DocumentImpl) node.getOwnerDocument())
                .getXmlFile();
        // avoid creation of cross file references in files with no name
        if (!parentFile.hasName && parentFile != nodeFile) {
            msgDis.popupMessage("No reference created");
            msgDis
                    .displayMessage("You cannot automatically create a cross file "
                            + "reference in a file with no name.\n"
                            + "Please, save this file first and retry.\n");
            return null;
        }
        // computes the href attribute of the new node, depending on the files
        String ref;
        if (!nodeFile.equals(parentFile)) {
            ref = nodeFile.getRelativePath(parentFile);
        } else {
            ref = "";
        }
        ref = ref + "#" + ((ElementImpl) node).getAttribute("name");
        return ref;
    }
}
