/*
 * com.simulogic.cese -- Cell Electrophysiology Simulation Environment (CESE)
 *
 * Copyright (C) 1999 - 2005 Sergey Missan
 * 
 * Dalhousie Univ., Halifax, NS, Canada
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.simulogic.cese;

import java.beans.Beans;
import java.io.IOException;
import java.net.URL;
import java.util.Iterator;
import java.util.Vector;
import java.util.jar.Attributes;
import java.util.jar.JarFile;

import javax.swing.beanbuilder.JarClassLoader;

/**
 * The ModelInstaller installs models from jar files (as JavaBeans)
 */
public class ModelInstaller extends Object {

    private Vector installedNames = new Vector();
    private Vector installedModels = new Vector();

    /**
     * Returns newly installed models names
     * 
     * @return installed model names
     */
    public Vector getInstalledNames() {
        return installedNames;
    }

    /**
     * Returns newly installed models
     * 
     * @return installed models
     */
    public Vector getInstalledModels() {
        return installedModels;
    }

    /**
     * Installs model(s) as the Java Beans in the jar file First it loads all
     * the Beans and resources into the JarClassLoader. Then it determines which
     * resources are beans (Java-Bean: True) from the Manifest.
     * 
     * @param filename full path to the jar file
     * @throws IOException
     */
    public void addJarFile(String filename) throws IOException {
        installedNames = new Vector();
        installedModels = new Vector();
        JarClassLoader loader = JarClassLoader.getJarClassLoader();
        if (loader == null) {
            // XXX - There has to be a better way of doing this!!
            // The problem is that there is no way to specify a null URL or
            // no arg JarClassLoader. The solution is to define a class loader
            // for the application.
            try {
                loader = new JarClassLoader(new URL("file:" + filename));
                JarClassLoader.setJarClassLoader(loader);
            } catch (Exception ex) {
                return;
            }
        } else {
            loader.addJarFile(filename);
        }

        JarFile jarfile;
        java.util.jar.Manifest mf;

        try {
            jarfile = new JarFile(filename);
            mf = jarfile.getManifest();
        } catch (IOException ex) {
            return;
        }

        Attributes attribs = mf.getMainAttributes();
   
        // Walk the list of Manifest attributes. The keys to the attributes
        // Are the names of the Jar entries. i.e.
        // sunw/demo/buttons/OurButton.class.
        Iterator iterator = mf.getEntries().keySet().iterator();
        while (iterator.hasNext()) {
            String beanName = (String) iterator.next();

            attribs = mf.getAttributes(beanName);

            if (attribs != null) {
                String isJavaBean = attribs
                        .getValue(JarClassLoader.Attributes_Name_JAVA_BEAN);

                if (isJavaBean != null && isJavaBean.equalsIgnoreCase("True")) {
                    addUserBean(loader, beanName);
                }
            }
        }

    }

    /**
     * Adds the bean to the user tab on the panel. This method determines if the
     * bean is from a class or from a serializable file.
     * 
     * @param loader ClassLoader to use for instantiation and class resolution
     * @param name Name of the bean and resource to use. The format should be
     *        the full path to the resource i.e.
     *        sunw/demo/buttons/OurButton.class.
     * @return loaded bean object
     * @throws IOException
     */
    public Object addUserBean(ClassLoader loader, String name)
            throws IOException {
        String classname = name;
        ModelInfo item = null;

        if (name.endsWith(".class")) {
            classname = name.substring(0, name.length() - 6);
        } else if (name.endsWith(".mdl")) {
            // Must deserialize the class.
            classname = name.substring(0, name.length() - 4);
        }
        classname = classname.replace('/', '.');

        try {
            Class cls = loader.loadClass(classname);
            item = new ModelInfo(cls);
        } catch (ClassNotFoundException ex) {
            // The object may be a serialized object. Instantiate it and
            // load it on the palette. This is for backwards compatibility
            // with serialized components on the BeanBox.
            try {
                Object obj = Beans.instantiate(loader, classname);
                item = new ModelInfo(obj, classname);
            } catch (Exception ex2) {
                System.out.println("addUserBean: Can't instantiate or load "
                        + classname);
            }
        }
        if (item != null) {
            try {
                if (((String) item.getBeanDescriptor().getValue("version"))
                        .equalsIgnoreCase(UserHome.CESE_VERSION_MAJOR)) {
                    installedNames.addElement(item.getDisplayName());
                    installedModels.addElement(item);
                }
            } catch (Exception e) {
                System.out.println("Invalid model version: "
                        + item.getDisplayName());
            }
        }

        return item.getModel();
    }

}