/*
 * com.simulogic.cese -- Cell Electrophysiology Simulation Environment (CESE)
 *
 * Copyright (C) 1999 - 2005 Sergey Missan
 * 
 * Dalhousie Univ., Halifax, NS, Canada
 *
 * BeanBuilder code Copyright (C) 2002 Mark Davidson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.simulogic.cese.properties;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.beans.PropertyEditorManager;

import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JToolBar;
import javax.swing.border.Border;

import com.simulogic.cese.ModelInfo;
import com.simulogic.cese.ResourceManager;

/**
 * The UI for listing, sorting, and setting component properties.
 */

public class PropertyPane extends JPanel {

    private JTable table;
    private PropertyColumnModel columnModel;
    private PropertyTableModel tableModel;

    private JToolBar viewBar;
    private JComboBox viewCombo;

    // View options.
    private static final String[] VIEW_CHOICES = {
            ResourceManager.getResource("All"),
            ResourceManager.getResource("Standard"),
            ResourceManager.getResource("Expert"),
            ResourceManager.getResource("Read_Only"),
            ResourceManager.getResource("Bound"),
            ResourceManager.getResource("Constrained"),
            ResourceManager.getResource("Hidden"),
            ResourceManager.getResource("Preferred")};

    /**
     * Constructor
     */
    public PropertyPane() {
        super(new BorderLayout());

        registerPropertyEditors();

        tableModel = new PropertyTableModel();
        columnModel = new PropertyColumnModel();
        table = new JTable(tableModel, columnModel);
        table.setAutoResizeMode(JTable.AUTO_RESIZE_LAST_COLUMN);

        add(createControlPanel(), BorderLayout.NORTH);
        add(new JScrollPane(table), BorderLayout.CENTER);
    }

    /**
     * Creates the panel for selecting the view and sorting order.
     */
    private JToolBar createControlPanel() {
        viewCombo = new JComboBox(VIEW_CHOICES);
        viewCombo
                .setSelectedItem(VIEW_CHOICES[PropertyTableModel.VIEW_STANDARD]);
        viewCombo.addActionListener(new java.awt.event.ActionListener() {

            public void actionPerformed(java.awt.event.ActionEvent evt) {
                Object obj = evt.getSource();

                if (obj instanceof JComboBox) {
                    if (obj == viewCombo) {
                        tableModel.filterTable(viewCombo.getSelectedIndex());
                    }
                }
            }
        });
        viewBar = new JToolBar();
        viewBar.add(viewCombo);

        return viewBar;
    }

    /**
     * Method which registers property editors for types.
     */
    private void registerPropertyEditors() {
        PropertyEditorManager.registerEditor(Color.class,
                javax.swing.beaninfo.SwingColorEditor.class);
        PropertyEditorManager.registerEditor(Font.class,
                javax.swing.beaninfo.SwingFontEditor.class);
        PropertyEditorManager.registerEditor(Border.class,
                javax.swing.beaninfo.SwingBorderEditor.class);
        PropertyEditorManager.registerEditor(Boolean.class,
                javax.swing.beaninfo.SwingBooleanEditor.class);
        PropertyEditorManager.registerEditor(boolean.class,
                javax.swing.beaninfo.SwingBooleanEditor.class);

        PropertyEditorManager.registerEditor(Integer.class,
                javax.swing.beaninfo.SwingIntegerEditor.class);
        PropertyEditorManager.registerEditor(int.class,
                javax.swing.beaninfo.SwingIntegerEditor.class);

        PropertyEditorManager.registerEditor(Double.class,
                javax.swing.beaninfo.SwingDoubleEditor.class);
        PropertyEditorManager.registerEditor(double.class,
                javax.swing.beaninfo.SwingDoubleEditor.class);

        PropertyEditorManager.registerEditor(Float.class,
                javax.swing.beaninfo.SwingNumberEditor.class);
        PropertyEditorManager.registerEditor(float.class,
                javax.swing.beaninfo.SwingNumberEditor.class);

        PropertyEditorManager.registerEditor(java.awt.Dimension.class,
                javax.swing.beaninfo.SwingDimensionEditor.class);
        PropertyEditorManager.registerEditor(java.awt.Rectangle.class,
                javax.swing.beaninfo.SwingRectangleEditor.class);
        PropertyEditorManager.registerEditor(java.awt.Insets.class,
                javax.swing.beaninfo.SwingInsetsEditor.class);

        PropertyEditorManager.registerEditor(String.class,
                javax.swing.beaninfo.SwingStringEditor.class);
        PropertyEditorManager.registerEditor(Object.class,
                javax.swing.beaninfo.SwingObjectEditor.class);
    }

    /**
     * Sets the PropertyPane to show the properties of the named bean.
     * 
     * @param info model info
     */
    public void setModelInfo(ModelInfo info) {
        if (info != null) {
            tableModel.setModelInfo(info);
            tableModel.filterTable(viewCombo.getSelectedIndex());
        }
    }

    /**
     * Clears table highlights
     */
    public void clearHighlights() {
    	tableModel.addChanges();
    	tableModel.clearChanges();
    }
    
    /**
	 * Updates table values
	 * 
	 * @param highlight
	 *            true if changes should be highlighted
	 */
	public void updateTable(boolean highlight) {
		tableModel.fireTableDataChanged();
		if (highlight)
			tableModel.addChanges();
	}

    /**
	 * Returns control toolbar
	 * 
	 * @return toolbar
	 */
    public JToolBar getToolBar() {
        return viewBar;
    }

    /**
     * Returns model for the property table
     * 
     * @return table model
     */
    public PropertyTableModel getTableModel() {
        return tableModel;
    }
}