/*
 * com.simulogic.dataview -- The Java2D Plotting Framework
 *
 * Copyright (C) 1999 - 2005 Sergey Missan
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.simulogic.dataview;

import java.awt.geom.Rectangle2D;

import javax.swing.tree.DefaultMutableTreeNode;

import com.simulogic.dataview.analysis.ChannelStats;

/**
 * Container for all the data (i.e. model)
 */
public class DataStorage extends javax.swing.tree.DefaultTreeModel {

	private double minx = Double.POSITIVE_INFINITY;

	private double maxx = Double.NEGATIVE_INFINITY;

	private double miny = Double.POSITIVE_INFINITY;

	private double maxy = Double.NEGATIVE_INFINITY;

	private Rectangle2D.Double limits;

	private DataChannel selectedChannel;

	private int c_count = 0;

	/**
	 * Default constructor
	 */
	public DataStorage() {
		super(new DefaultMutableTreeNode(ResourceManager.getResource("Legend")));

		limits = new Rectangle2D.Double(Double.POSITIVE_INFINITY,
				Double.NEGATIVE_INFINITY, 0, 0);
	}

	/**
	 * Adds a group to the storage
	 * 
	 * @param g
	 *            a group to add
	 */
	public synchronized void addGroup(DataGroup g) {
		((DefaultMutableTreeNode) root).add(g);
		int i = getGroupsSize() - 1;

		setLimits(i);
		for (int j = 1; j < getChannelsSize(i); j++) {
			getChannel(i, j).getAttribute().setIndexedColor(c_count);
			getChannel(i, j).getAttribute().setIndexedFillColor(c_count);
			getChannel(i, j).getAttribute().setIndexedSymbol(c_count);
			c_count++;
		}

		int[] ind = new int[1];
		ind[0] = getGroupsSize() - 1;
		nodesWereInserted(root, ind);

		ind = new int[1];
		ind[0] = getChannelsSize(i) - 1;
		nodesWereInserted(root.getChildAt(i), ind);
	}

	/**
	 * Adds a channel to the specific group in the storage
	 * 
	 * @param c
	 *            channel to add
	 * @param g
	 *            target group
	 */
	public synchronized void addChannel(DataChannel c, int g) {
		((DataGroup) root.getChildAt(g)).addChannel(c);

		setLimits(g);
		c.getAttribute().setIndexedColor(c_count);
		c.getAttribute().setIndexedFillColor(c_count);
		c.getAttribute().setIndexedSymbol(c_count);
		c_count++;

		int[] ind = new int[1];
		ind[0] = getChannelsSize(g) - 1;
		nodesWereInserted(root.getChildAt(g), ind);
	}

	/**
	 * Returns group from this storage
	 * 
	 * @param g
	 *            group index
	 * @throws ClassCastException
	 *             if object is not of type DataGroup
	 * @return DataGroup for index
	 */
	public DataGroup getGroup(int g) throws ClassCastException {
		return (DataGroup) root.getChildAt(g);
	}

	/**
	 * Returns number of groups in this storage
	 * 
	 * @return number of groups
	 */
	public int getGroupsSize() {
		return root.getChildCount();
	}

	/**
	 * Returns channel in the group from this storage
	 * 
	 * @return DataChannel for index
	 * @param c
	 *            channel index in group g
	 * @param g
	 *            group index
	 * @throws ClassCastException
	 *             if object is not of type DataChannel
	 */
	public DataChannel getChannel(int g, int c) throws ClassCastException {
		return ((DataGroup) root.getChildAt(g)).getChannel(c);
	}

	/**
	 * Returns node that keeps channel in the group from this storage
	 * 
	 * @return DataChannel for index
	 * @param c
	 *            channel index in group g
	 * @param g
	 *            group index
	 * @throws ClassCastException
	 *             if object is not of type DefaultMutableTreeNode
	 */
	public DefaultMutableTreeNode getChannelNode(int g, int c)
			throws ClassCastException {
		return (DefaultMutableTreeNode) root.getChildAt(g).getChildAt(c);
	}

	/**
	 * Returns number of channels in the specific group
	 * 
	 * @param g
	 *            target group
	 * @return number of channels in group
	 */
	public int getChannelsSize(int g) {
		return root.getChildAt(g).getChildCount();
	}

	/**
	 * Returns X channel that corresponds to the given channel
	 * 
	 * @param chan
	 *            channel to test
	 * @return X channel in the correspondent group
	 */
	public DataChannel getXChannel(DataChannel chan) {
		try {
			for (int i = 0; i < getGroupsSize(); i++) {
				DataGroup group = getGroup(i);
				DataChannel xchannel = getChannel(i, group.getXChannel());

				for (int j = 0; j < getChannelsSize(i); j++) {
					DataChannel channel = getChannel(i, j);

					if (channel == chan) {
						return xchannel;
					}
				}
			}
		} catch (Exception e) {
		}
		return null;
	}

	/**
	 * Returns true if given channel is an X channel in one of the groups
	 * 
	 * @param chan
	 *            channel to test
	 * @return true if X channel
	 */
	public boolean isXChannel(DataChannel chan) {
		try {
			for (int i = 0; i < getGroupsSize(); i++) {
				DataGroup group = getGroup(i);
				DataChannel xchannel = getChannel(i, group.getXChannel());

				if (xchannel == chan) {
					return true;
				}
			}
		} catch (Exception e) {
		}
		return false;
	}

	/**
	 * Sets an attribute of the given channel to X channel and adjusts all
	 * entries in the correspondent group
	 * 
	 * @param chan
	 *            channel to test
	 */
	public void setXChannel(DataChannel chan) {
		try {
			for (int i = 0; i < getGroupsSize(); i++) {
				DataGroup group = getGroup(i);
				for (int j = 0; j < getChannelsSize(i); j++) {
					DataChannel xchannel = getChannel(i, j);

					if (xchannel == chan) {
						group.setXChannel(j);
					}
				}
			}
		} catch (Exception e) {
		}
	}

	/**
	 * Returns a currently selected channel
	 * 
	 * @return selected channel
	 */
	public DataChannel getSelectedChannel() {
		return selectedChannel;
	}

	/**
	 * Sets the new selected channel
	 * 
	 * @param selectedChannel
	 *            selected channel
	 */
	public void setSelectedChannel(DataChannel selectedChannel) {
		this.selectedChannel = selectedChannel;
	}

	/**
	 * Returns rectangle of data limits
	 * 
	 * @return limit rectangle
	 */
	public Rectangle2D.Double getLimitRectangle() {
		return limits;
	}

	/**
	 * Returns autoscale rectangle for data interval
	 * 
	 * @param xMin
	 *            minimum limit for x
	 * @param xMax
	 *            maximum limit for x
	 * @return autoscale rectangle
	 */

	public Rectangle2D.Double getAutoscaleRectangle(double xMin, double xMax) {
		com.simulogic.dataview.DataChannel xchannel = null;
		com.simulogic.dataview.DataChannel ychannel = null;

		Rectangle2D.Double bounds = new Rectangle2D.Double(xMin,
				Double.NEGATIVE_INFINITY, Math.abs(xMax - xMin), 0);

		double minpeak = Double.POSITIVE_INFINITY;
		double maxpeak = Double.NEGATIVE_INFINITY;

		try {
			for (int i = 0; i < getGroupsSize(); i++) {
				DataGroup group = getGroup(i);
				xchannel = getChannel(i, group.getXChannel());

				for (int j = 0; j < getChannelsSize(i); j++) {
					ychannel = getChannel(i, j);

					if ((xchannel != null) && (ychannel != null)
							&& (ychannel.getAttribute().isNormal())
							&& (ychannel != xchannel)) {
						int k = ChannelStats.getNegativePeakIndex(xMin, xMax,
								xchannel, ychannel);

						if ((k < xchannel.size()) && (k >= 0))
							minpeak = Math.min(minpeak, ychannel.getData(k));

						k = ChannelStats.getPositivePeakIndex(xMin, xMax,
								xchannel, ychannel);

						if ((k < xchannel.size()) && (k >= 0))
							maxpeak = Math.max(maxpeak, ychannel.getData(k));
					}
				}
			}
		} catch (Exception e) {
		}
		bounds.y = maxpeak;
		bounds.height = Math.abs(maxpeak - minpeak);
		return bounds;
	}

	/**
	 * Updates group in the storage
	 * 
	 * @param g
	 *            target group index
	 */
	public synchronized void updateModel(int g) {
		setLimits(g);
		nodeChanged(root.getChildAt(g));
	}

	private void setLimits(int i) {
		DataGroup group = getGroup(i);

		for (int j = 0; j < getChannelsSize(i); j++) {
			DataChannel channel = getChannel(i, j);

			if (j == group.getXChannel()) {
				minx = Math.min(minx, channel.minLimit());
				maxx = Math.max(maxx, channel.maxLimit());
			} else {
				miny = Math.min(miny, channel.minLimit());
				maxy = Math.max(maxy, channel.maxLimit());
			}
		}

		limits
				.setRect(minx, maxy, Math.abs(maxx - minx), Math.abs(maxy
						- miny));
	}

}