/*
 * com.simulogic.dataview -- The Java2D Plotting Framework
 * 
 * Copyright (C) 1999 - 2005 Sergey Missan
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this library; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * MThumbSlider: Copyright (C) 1999 Nobuo Tamemasa.
 *  
 */

package com.simulogic.dataview.cursors;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.util.Arrays;

import javax.swing.Icon;
import javax.swing.JComponent;
import javax.swing.JSlider;
import javax.swing.SwingConstants;
import javax.swing.UIDefaults;
import javax.swing.UIManager;
import javax.swing.event.ChangeListener;
import javax.swing.event.MouseInputAdapter;
import javax.swing.plaf.ColorUIResource;
import javax.swing.plaf.ComponentUI;
import javax.swing.plaf.basic.BasicSliderUI;
import javax.swing.plaf.metal.MetalIconFactory;

public class BasicCursorSliderUI extends BasicSliderUI implements
		CursorSliderAdditional {

	CursorSliderAdditionalUI additionalUI;

	MouseInputAdapter cursorTrackListener;

	ChangeListener cursorChangeListener;

	Icon thumbRenderer;

	public static ComponentUI createUI(JComponent c) {
		return new BasicCursorSliderUI((JSlider) c);
	}

	public BasicCursorSliderUI(JSlider b) {
		super(b);
	}

	public void installUI(JComponent c) {
		additionalUI = new CursorSliderAdditionalUI(this);
		additionalUI.installUI(c);
		cursorTrackListener = createCursorTrackListener((JSlider) c);
		cursorChangeListener = createCursorChangeListener((JSlider) c);
		super.installUI(c);

		UIDefaults defaults = UIManager.getDefaults();
		if (slider.getOrientation() == SwingConstants.HORIZONTAL) {
			thumbRenderer = MetalIconFactory.getHorizontalSliderThumbIcon();
			defaults.put("Slider.horizontalThumbIcon", thumbRenderer);
		} else {
			thumbRenderer = MetalIconFactory.getVerticalSliderThumbIcon();
			defaults.put("Slider.verticalThumbIcon", thumbRenderer);
		}

		java.util.List sliderGradient = Arrays.asList(new Object[] {
				new Float(.3f), new Float(.2f),
				new ColorUIResource(defaults.getColor("Slider.foreground")),
				new ColorUIResource(0xFFFFFF),
				new ColorUIResource(UIManager.getColor("Slider.background")) });

		defaults.put("Slider.gradient", sliderGradient);
		defaults.put("Slider.focusGradient", sliderGradient);
	}

	public void uninstallUI(JComponent c) {
		super.uninstallUI(c);
		additionalUI.uninstallUI(c);
		additionalUI = null;
		cursorTrackListener = null;
		cursorChangeListener = null;
	}

	protected MouseInputAdapter createCursorTrackListener(JSlider slider) {
		return additionalUI.trackListener;
	}

	protected TrackListener createTrackListener(JSlider slider) {
		return null;
	}

	protected ChangeListener createCursorChangeListener(JSlider slider) {
		return additionalUI.changeHandler;
	}

	protected void installListeners(JSlider slider) {
		slider.addMouseListener(cursorTrackListener);
		slider.addMouseMotionListener(cursorTrackListener);
		slider.addFocusListener(focusListener);
		slider.addComponentListener(componentListener);
		slider.addPropertyChangeListener(propertyChangeListener);
		slider.addChangeListener(cursorChangeListener);
	}

	protected void uninstallListeners(JSlider slider) {
		slider.removeMouseListener(cursorTrackListener);
		slider.removeMouseMotionListener(cursorTrackListener);
		slider.removeFocusListener(focusListener);
		slider.removeComponentListener(componentListener);
		slider.removePropertyChangeListener(propertyChangeListener);
		slider.removeChangeListener(cursorChangeListener);
	}

	protected void calculateGeometry() {
		super.calculateGeometry();
		additionalUI.calculateThumbSize();
		additionalUI.calculateThumbLocation();
	}

	protected void calculateThumbLocation() {
		additionalUI.calculateThumbLocation();
	}

	public void paint(Graphics g, JComponent c) {
		Rectangle clip = g.getClipBounds();
		Rectangle[] thumbRects = additionalUI.getThumbRects();
		thumbRect = thumbRects[0];
		int thumbNum = additionalUI.getThumbNum();

		if (slider.getPaintTrack() && clip.intersects(trackRect))
			paintTrack(g);

		for (int i = thumbNum - 1; 0 <= i; i--) {
			if (clip.intersects(thumbRects[i])) {
				thumbRect = thumbRects[i];

				paintThumb(g, i);
			}
		}
	}

	public void paintThumb(Graphics g, int i) {
		Graphics2D g2 = (Graphics2D) g;
		Point2D.Double dst = new Point2D.Double();
		int y = (int) ((CursorSlider) slider).getView().getScrollPane()
				.getViewport().getViewPosition().getY();
		int e = (int) ((CursorSlider) slider).getView().getScrollPane()
				.getViewport().getExtentSize().getHeight();

		g2.setColor(UIManager.getColor("Slider.foreground"));
		g2.setStroke(new java.awt.BasicStroke(1.0f));

		if (slider.getOrientation() == SwingConstants.HORIZONTAL) {
			// vertical
			g2.drawLine(thumbRect.x + thumbRect.width / 2, y, thumbRect.x
					+ thumbRect.width / 2, y + e);

			// horizontal
			com.simulogic.dataview.DataChannel s = ((CursorSlider) slider)
					.getView().getStorage().getSelectedChannel();

			if ((((CursorSlider) slider).isLocked())
					&& (s != null)
					&& (s.getAttribute().isNormal())
					&& (!((CursorSlider) slider).getView().getStorage()
							.isXChannel(s))) {
				double tst = com.simulogic.dataview.analysis.ChannelStats
						.getValueAtX(((CursorSlider) slider).getValueAt(i),
								((CursorSlider) slider).getView().getStorage()
										.getXChannel(s), s);

				((CursorSlider) slider).getView().getTransform().transform(
						new Point2D.Double(0.0, tst), dst);
				g2.drawLine(thumbRect.x, thumbRect.y + (int) dst.getY(),
						thumbRect.x + thumbRect.width - 1, thumbRect.y
								+ (int) dst.getY());
			} else {
				g2.drawLine(thumbRect.x, thumbRect.y + y + e / 2, thumbRect.x
						+ thumbRect.width - 1, thumbRect.y + y + e / 2);
			}

			// label
			g2.drawString("" + (i + 1), thumbRect.x + thumbRect.width / 2 + 1,
					thumbRect.y
							+ thumbRenderer.getIconHeight()
							+ slider.getFontMetrics(slider.getFont())
									.getHeight() + y);
		}
		thumbRenderer.paintIcon(slider, g2, thumbRect.x, thumbRect.y + y);
	}

	public void fillTrack(Graphics g, Point t1, Point t2, Color fillColor) {
	}

	public void scrollByBlock(int direction) {
	}

	public void scrollByUnit(int direction) {
	}

	//
	// CursorSliderAdditional
	//
	public Rectangle getTrackRect() {
		trackRect = new Rectangle(thumbRect.width / 2, 0,
				((CursorSlider) slider).getView().getPreferredSize().width,
				((CursorSlider) slider).getView().getPreferredSize().height);
		return trackRect;
	}

	public Rectangle[] getThumbRects() {
		return additionalUI.getThumbRects();
	}

	protected int getThumbOverhang() {
		return 0;
	}

	public Dimension getThumbSize() {
		Dimension size = new Dimension();

		if (thumbRenderer != null) {
			size.width = thumbRenderer.getIconWidth();
			size.height = thumbRenderer.getIconHeight();
		}
		return size;
	}

	public double xPositionForValue(double value) {
		return additionalUI.xPositionForValue(value);
	}

	public double yPositionForValue(double value) {
		return additionalUI.yPositionForValue(value);
	}

}